import time
import threading
import os
import json
import ssl
import uuid
from typing import Callable
import paho.mqtt.client as mqtt
import logging

cert_dir = f"/etc/vyomcloudbridge/mosquito/certs/"

# /etc/vyomcloudbridge/mosquito/certs/machine.cert.pem
cert_file_name = "machine.cert.pem"
cert_file_path = os.path.join(cert_dir, cert_file_name)

# /etc/vyomcloudbridge/mosquito/certs/machine.private.key
pri_key_file_name = "machine.private.key"
pri_key_file_path = os.path.join(cert_dir, pri_key_file_name)

# /etc/vyomcloudbridge/mosquito/certs/root-CA.crt
root_ca_file_name = "root-CA.crt"
root_ca_file_path = os.path.join(cert_dir, root_ca_file_name)

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


class MqttDeviceClient:
    def __init__(
        self,
        broker_host: str,
        broker_port: int = 8883,  # SSL/TLS port
        callback: Callable = None,
    ):
        try:
            self.machine_id = "32"
            self.broker_host = broker_host
            self.broker_port = broker_port
            self.callback = callback
            self.client_id = f"device_{self.machine_id}_{uuid.uuid4().hex[:8]}"

            # Certificate paths
            self.cert_path = cert_file_path
            self.pri_key_path = pri_key_file_path
            self.root_ca_path = root_ca_file_path

            self._verify_cert_files()

            # Connection state tracking
            self.is_connected = False
            self.connection_lock = threading.Lock()
            self._connection_in_progress = False

            # Subscription tracking
            self.subscribed_topics = set()

            # Connection and reconnection parameters
            self.max_reconnect_attempts = 5
            self.base_reconnect_delay = 1
            self.reconnect_delay_max = 60

            # Device topic prefix (restricted to device-specific topics)
            self.device_topic_prefix = f"device/{self.machine_id}"

            # Initialize MQTT client
            self.mqtt_client = None
            self._create_mqtt_client()

            # Start connection monitoring
            self._start_connection_monitor()

        except Exception as e:
            logger.error(f"Error initializing MqttDeviceClient: {str(e)}")
            raise

    def _verify_cert_files(self):
        """Verify that all required certificate files exist"""
        for file_path in [self.cert_path, self.pri_key_path, self.root_ca_path]:
            if not os.path.exists(file_path):
                logger.error(f"Certificate file not found: {file_path}")
                raise FileNotFoundError(f"Required certificate file not found: {file_path}")

    def _create_mqtt_client(self):
        """Create and configure MQTT client with SSL/TLS"""
        try:
            # Create MQTT client
            self.mqtt_client = mqtt.Client(
                client_id=self.client_id,
                clean_session=False,
                protocol=mqtt.MQTTv311
            )

            # Configure SSL/TLS with certificates
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            context.check_hostname = False
            context.verify_mode = ssl.CERT_REQUIRED
            
            # Load certificates
            context.load_verify_locations(self.root_ca_path)
            context.load_cert_chain(self.cert_path, self.pri_key_path)

            self.mqtt_client.tls_set_context(context)

            # Set callbacks
            self.mqtt_client.on_connect = self._on_connect
            self.mqtt_client.on_disconnect = self._on_disconnect
            self.mqtt_client.on_message = self._on_message
            self.mqtt_client.on_subscribe = self._on_subscribe
            self.mqtt_client.on_publish = self._on_publish

            # Configure keep alive and other options
            self.mqtt_client.max_inflight_messages_set(20)
            self.mqtt_client.max_queued_messages_set(100)

            logger.info("MQTT client created successfully with SSL/TLS configuration")

        except Exception as e:
            logger.error(f"Failed to create MQTT client: {str(e)}")
            raise

    def _on_connect(self, client, userdata, flags, rc):
        """Callback for when client connects to broker"""
        if rc == 0:
            logger.info(f"Device client connected successfully to {self.broker_host}:{self.broker_port}")
            with self.connection_lock:
                self.is_connected = True
                self._connection_in_progress = False
            
            # Resubscribe to topics
            self._resubscribe_to_topics()
        else:
            logger.error(f"Failed to connect, return code: {rc}")
            with self.connection_lock:
                self.is_connected = False
                self._connection_in_progress = False

    def _on_disconnect(self, client, userdata, rc):
        """Callback for when client disconnects from broker"""
        logger.warning(f"Device client disconnected, return code: {rc}")
        with self.connection_lock:
            self.is_connected = False

    def _on_message(self, client, userdata, msg):
        """Callback for when message is received"""
        try:
            topic = msg.topic
            payload = msg.payload.decode('utf-8')
            logger.info(f"Received message from topic '{topic}': {payload}")
            
            if self.callback:
                self.callback(topic, payload)
            else:
                logger.info("No callback provided, skipping callback execution")
                
        except Exception as e:
            logger.error(f"Error processing received message: {str(e)}")

    def _on_subscribe(self, client, userdata, mid, granted_qos):
        """Callback for when subscription is confirmed"""
        logger.info(f"Subscription confirmed with QoS: {granted_qos}")

    def _on_publish(self, client, userdata, mid):
        """Callback for when message is published"""
        logger.debug(f"Message published with mid: {mid}")

    def connect(self):
        """Connect to MQTT broker with exponential backoff"""
        with self.connection_lock:
            if self.is_connected:
                logger.info("Already connected to MQTT broker")
                return True
            
            if self._connection_in_progress:
                logger.info("Connection attempt already in progress")
                return False
            
            self._connection_in_progress = True

        try:
            for attempt in range(self.max_reconnect_attempts):
                try:
                    delay = min(self.base_reconnect_delay * (2 ** attempt), self.reconnect_delay_max)
                    
                    if attempt > 0:
                        logger.info(f"Reconnection attempt {attempt + 1}/{self.max_reconnect_attempts}")
                        time.sleep(delay)

                    # Connect to broker
                    self.mqtt_client.connect(self.broker_host, self.broker_port, 60)
                    self.mqtt_client.loop_start()
                    
                    # Wait for connection
                    timeout = 10
                    start_time = time.time()
                    while not self.is_connected and (time.time() - start_time) < timeout:
                        time.sleep(0.1)
                    
                    if self.is_connected:
                        logger.info("Successfully connected to MQTT broker")
                        return True
                    else:
                        logger.warning(f"Connection attempt {attempt + 1} timed out")
                        
                except Exception as e:
                    logger.error(f"Connection attempt {attempt + 1} failed: {str(e)}")
            
            logger.error("Failed to connect after all attempts")
            return False
            
        finally:
            with self.connection_lock:
                self._connection_in_progress = False

    def _resubscribe_to_topics(self):
        """Resubscribe to all previously subscribed topics"""
        for topic in self.subscribed_topics.copy():
            try:
                result = self.mqtt_client.subscribe(topic, qos=1)
                if result[0] == mqtt.MQTT_ERR_SUCCESS:
                    logger.info(f"Resubscribed to topic: {topic}")
                else:
                    logger.error(f"Failed to resubscribe to topic: {topic}")
            except Exception as e:
                logger.error(f"Error resubscribing to {topic}: {str(e)}")

    def subscribe_to_topic(self, topic: str):
        """Subscribe to a topic (restricted to device topics)"""
        try:
            # Enforce topic restrictions for device clients
            if not self._is_topic_allowed(topic):
                logger.error(f"Topic '{topic}' is not allowed for device client")
                raise PermissionError(f"Device client not allowed to subscribe to topic: {topic}")

            if not self.is_connected:
                if not self.connect():
                    raise ConnectionError("Failed to connect to MQTT broker")

            result = self.mqtt_client.subscribe(topic, qos=1)
            if result[0] == mqtt.MQTT_ERR_SUCCESS:
                self.subscribed_topics.add(topic)
                logger.info(f"Subscribed to topic: {topic}")
            else:
                logger.error(f"Failed to subscribe to topic: {topic}")
                raise Exception(f"Subscription failed with code: {result[0]}")

        except Exception as e:
            logger.error(f"Subscription to {topic} failed: {str(e)}")
            raise

    def publish_message(self, topic: str, payload, retain: bool = False):
        """Publish message (restricted to device topics)"""
        try:
            # Enforce topic restrictions for device clients
            if not self._is_topic_allowed(topic):
                logger.error(f"Topic '{topic}' is not allowed for device client")
                return False

            if not self.is_connected:
                if not self.connect():
                    logger.error("Failed to connect to MQTT broker for publishing")
                    return False

            # Convert payload to string if it's a dict
            if isinstance(payload, dict):
                payload = json.dumps(payload)

            logger.info(f"Publishing message to topic: {topic}")
            result = self.mqtt_client.publish(topic, payload, qos=1, retain=retain)
            
            if result.rc == mqtt.MQTT_ERR_SUCCESS:
                logger.info(f"Message published successfully to topic: {topic}")
                return True
            else:
                logger.error(f"Failed to publish message, return code: {result.rc}")
                return False

        except Exception as e:
            logger.error(f"Publish to {topic} failed: {str(e)}")
            return False

    def _is_topic_allowed(self, topic: str) -> bool:
        """Check if topic is allowed for device client"""
        allowed_patterns = [
            self.device_topic_prefix,  # device/machine_id/*
            f"commands/{self.machine_id}",  # commands/machine_id/*
            f"config/{self.machine_id}",   # config/machine_id/*
        ]
        
        return any(topic.startswith(pattern) for pattern in allowed_patterns)

    def _start_connection_monitor(self):
        """Start background thread to monitor connection"""
        def monitor_connection():
            while True:
                try:
                    if not self.is_connected and not self._connection_in_progress:
                        logger.warning("Connection lost. Attempting to reconnect...")
                        self.connect()
                    time.sleep(30)  # Check every 30 seconds
                except Exception as e:
                    logger.error(f"Connection monitoring failed: {str(e)}")
                    time.sleep(30)

        monitor_thread = threading.Thread(target=monitor_connection, daemon=True)
        monitor_thread.start()

    def disconnect(self):
        """Gracefully disconnect from MQTT broker"""
        try:
            if self.mqtt_client:
                self.mqtt_client.loop_stop()
                self.mqtt_client.disconnect()
                
            with self.connection_lock:
                self.is_connected = False
                
            logger.info("MQTT device client disconnected successfully")
            
        except Exception as e:
            logger.error(f"Failed to disconnect MQTT client: {str(e)}")

    def get_allowed_topics(self):
        """Get list of topics this device client is allowed to access"""
        return [
            f"{self.device_topic_prefix}/*",
            f"commands/{self.machine_id}/*",
            f"config/{self.machine_id}/*"
        ]
    

def message_callback(topic, payload):
    try:
        logger.info(f"Received message in callback '{topic}': {json.loads(payload)}")
    except Exception as e:
        logger.error("Error in calling callback")


def main():
    """Example usage of MosquittoMqttClient"""
    machine_id = "32"
    # Initialize client
    client = MqttDeviceClient(
        broker_host="localhost",  # Change to your EC2 instance IP for remote access
        broker_port=1883,
        machine_id="32",
        callback=message_callback,
        # username="your_username",  # Uncomment if authentication is enabled
        # password="your_password",  # Uncomment if authentication is enabled
    )
    
    # Wait for connection
    time.sleep(2)
    subscribe_topic_1 = f"vyom-mqtt-msg/{machine_id}/"
    subscribe_topic_2 = f"vyom-mqtt-msg/{machine_id}/#"
    
    
    # Subscribe to topics
    client.subscribe_to_topic(subscribe_topic_1)
    client.subscribe_to_topic(subscribe_topic_2)
    
    publish_topic_1 = f"1/2025-02-11/33/44/{machine_id}/99/10/hello.json"
    # Publish some test messages
    client.publish_message(publish_topic_1, "Hello from Mosquitto client!")
    client.publish_message(publish_topic_1, {"message": "JSON payload", "timestamp": time.time()})
    
    # Keep the client running
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Shutting down...")
        client.close_connection()


if __name__ == "__main__":
    main()